; ChiWriter Screen Driver
; Copyright (C) 1988 Horstmann Software. All Rights Reserved.
;
; To make this into a driver file, assemble it--
; 	MASM -DscreenType -Dmode [-DINV] SCREEN.ASM;
; e.g. 
;	MASM -DCGA -DLO SCREEN.ASM;
; or
;   	MASM -DEGA -DHI -DINV SCREEN.ASM
; link it--
;	LINK SCREEN;
; ignore the "missing stack segment" and convert into a .COM file--
; 	EXE2BIN SCREEN.EXE <driver name>.DRV
;
; Add a -DINV to get an inverse display. (Black characters on white background)
; NOTE: The inverse display doesn't work well with the font designer
;
; Supported graphics boards are
;
; CGA		-DCGA
; EGA		-DEGA
; VGA		-DVGA
; Hercules	-DHERC
; Tecmar	-DTEC
; AT&T 6300	-DATT
; Toshiba 	-DTOSH
; Wyse 700	-DWYSE
; Sigma 400	-DSIG
; Tandy 2000	-DTAN
;
;       VGAHI is EGA HI 
;       VGALO is VGA HI (!)
;
; Comments on graphics boards:
; 
; (1) CGA
; No, you cannot change the colors on CGA. 
; The CGA driver is -DLO, there seems to be no point in a CGA HI driver.
;
; (2) VGA
; VGAHI is the same as EGAHI. It just uses the card's EGA emulation.
; VGALO is actually compiled with -DVGA -DHI. Due to the 480 vertical pixels,
; the letters come out quite small already and -DVGA -DLO looks awful.
;
; (3) Wyse 700
; The drivers on the support disk use the 640 x 400 mode. 
; Assembling with -DWYSE -DXHI selects the WYSE 700 1280 x 800 display mode.
; Use the .PFT fonts (16 x 24 box) in that mode.
; You should also select -DINV, otherwise the characters are very hard to read.
; The font designer does not work in this mode.
;   
; (4) Tecmar graphics board
; The Tecmar driver has been adapted by Prof. Leo Jonker from Queen's 
; University, Kingson, Ontario. He gives the following comments:
;      Chiwriter uses the first half of segment B000H as video buffer, and
; writes to the MONOCHROME screen in the 720 x 352 pixel resolution 2 -
; intensity graphics mode described in the TECMAR manual.  To operate in
; this mode the jumper settings on the TECMAR board should be as follows:
; 
;        JPR1:   Position B must be on.  The other positions may be on
;                or off depending on other software you may wish to run.
;        JPR4:   On.
;        JPR5:   Position B.
;        JPR6:   On.
;        JPR7:   On.
;        JPR8:   Off.;
;
; (5) Sigma 400
; The Sigma driver has been adapted by 	Timothy Williams, Physics Department
; College of William and Mary, Williamsburg, VA 23185. He writes:
; The driver writes to only one of the four memory planes on the Sigma board, 
; which turns out to be the green plane; this means that ChiWriter runs 
; in green.
;
;---------------------------------------------------------------------------
;
; The following services are provided:
;
;       Initialize graphics board. Switch to text mode & clear screen
;
; AH = 0
;
;       Terminate. Reset graphics board for return to DOS
;
; AH = 1
;
;       Switch to text mode and clear screen
;
; AH = 2
;
;       Switch to graphics mode and clear screen
;
; AH = 3
;
;       Put a bit pattern on the screen
;
; AH = 4
; BX = pixel row of pattern 
; CX = leftmost pixel col of character
; DX = length of pattern in bit
; DS:SI = address of char
;
;       Or a character on the edit screen
;
; AH = 5
; BX = bottom edit row of char
; DI = edit col of char
; DX = top row of edit screen that is permitted to change
; CX = bottom row of edit screen that is allowed to change
; DS:SI = address of char 
;
;       Xor a character on the edit screen
;
; AH = 6
; BX = bottom edit row of char
; DI = edit col of char
; DX = top row of edit screen that is permitted to change
; CX = bottom row of edit screen that is allowed to change
; DS:SI = address of char 
;
;       Delete an area on the edit screen
;
; AH = 7
; BH = r1
; BL = r2
; CH = c1
; CL = c2
;
;       Move an area on the edit screen
;
; AH = 8
; BH = r1
; BL = r2
; CH = c1
; CL = c2
; DH = dr
; DL = dc
;
; Note that each character is 8 bit wide and 16 scan rows high.
; Each ChiWriter row is 8 scan rows high.
;
;
; Screen memory map example 
; SCAN_BLKS = 4, BLK_DIFF = 2000H
;
; +--------------------------------+
; | xxxxxxxxxxx unused xxxxxxxxxxx |
; |				   |
; |   scan lines 3, 7, ... 399     |
; +--------------------------------+ <--- 6000H
; | xxxxxxxxxxx unused xxxxxxxxxxx |
; |				   |
; |   scan lines 2, 6, ... 398     |
; +--------------------------------+ <--- 4000H
; | xxxxxxxxxxx unused xxxxxxxxxxx |
; |				   |
; |   scan lines 1, 5, ... 397     |
; +--------------------------------+ <--- 2000H
; | xxxxxxxxxxx unused xxxxxxxxxxx |
; |				   |
; |   scan lines 0, 4, ... 396     |
; +--------------------------------+ <--- 0000H
;
;****************************************************************************
; Equates
;****************************************************************************
;
;
ifdef CGA
SC_VSIZE EQU  	200	; vertical screen size in pixels
VIDEO_SEG EQU 	0B800H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	2 	; the number of scan line blocks
GRMODE	EQU	6	; BIOS graphics mode
endif

ifdef VGA
SC_VSIZE EQU  	480	; vertical screen size in pixels
VIDEO_SEG EQU 	0A000H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	0	; the difference between successive scan line blocks
SCAN_BLKS EQU 	1 	; the number of scan line blocks
GRMODE	EQU	11H	; BIOS graphics mode
endif

ifdef EGA
SC_VSIZE EQU  	480	; vertical screen size in pixels
VIDEO_SEG EQU 	0A000H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	0	; the difference between successive scan line blocks
SCAN_BLKS EQU 	1 	; the number of scan line blocks
endif

ifdef HERC
SC_VSIZE EQU  	348	; vertical screen size in pixels
VIDEO_SEG EQU 	0B000H 	; start of video segment
SCAN_DIFF EQU 	90	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	4 	; the number of scan line blocks
endif

ifdef TEC
SC_VSIZE EQU  	348	; vertical screen size in pixels
VIDEO_SEG EQU 	0B000H 	; start of video segment
SCAN_DIFF EQU 	90	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	4 	; the number of scan line blocks
endif

ifdef ATT
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0B800H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	4 	; the number of scan line blocks
GRMODE	EQU	40H	; BIOS graphics mode
endif

ifdef TOSH
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0B800H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	4 	; the number of scan line blocks
GRMODE	EQU	74H	; BIOS graphics mode
endif

ifdef SIG
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0B800H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	2000H	; the difference between successive scan line blocks
SCAN_BLKS EQU 	4 	; the number of scan line blocks
GRMODE	EQU	42H	; BIOS graphics mode
endif

ifdef TAN
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0E000H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	0	; the difference between successive scan line blocks
SCAN_BLKS EQU 	1 	; the number of scan line blocks
GRMODE	EQU	9	; BIOS graphics mode
endif

ifdef WYSE

ifdef XHI
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0A000H 	; start of video segment
SCAN_DIFF EQU 	160	; the distance between successive scan lines
BLK_DIFF  EQU  	0	; the difference between successive scan line blocks
SCAN_BLKS EQU 	1 	; the number of scan line blocks
else
SC_VSIZE EQU  	400	; vertical screen size in pixels
VIDEO_SEG EQU 	0A000H 	; start of video segment
SCAN_DIFF EQU 	80	; the distance between successive scan lines
BLK_DIFF  EQU  	0	; the difference between successive scan line blocks
SCAN_BLKS EQU 	1 	; the number of scan line blocks
endif
endif

;
ifdef HI
CH_VSIZE EQU	8	; vertical half-char box size in pixels
endif
ifdef LO
CH_VSIZE EQU	5	; vertical half-char box size in pixels
endif
ifdef XHI
CH_VSIZE EQU	6	; vertical half-char box size in pixels
endif

ifdef XHI
CH_HBYTES EQU  	2	; horizontal char box size in bytes
SC_HSIZE EQU  	1280	; horizontal screen size in pixels
else
CH_HBYTES EQU  	1	; horizontal char box size in bytes
SC_HSIZE EQU  	640	; horizontal screen size in pixels
endif

MAX_HPOS EQU  	SC_HSIZE/(8*CH_HBYTES) - 1	; maximal horizontal position
MAX_VPOS EQU	SC_VSIZE/CH_VSIZE - 1		; maximal vertical position

;****************************************************************************
; The offset corresponding to pixel coordinates (r,c) is
; 	offset = (r/SCAN_BLKS) * SCAN_DIFF + (r%SCAN_BLKS) * BLK_DIFF
;
; NOTE: If SCAN_BLKS divides SCAN_DIFF, this is computed as follows:
; 	mod = r%SCAN_BLKS;
; 	offset = (r - mod) * (SCAN_DIFF/SCAN_BLKS) + mod * SCAN_BLKS;
;
;****************************************************************************
; MACROS
;****************************************************************************
if1
shiftdiv MACRO reg,pow2
	n = 1
	REPT	pow2	; actually less often
if n eq pow2
	EXITM
endif
if n gt pow2
	%OUT	*** Error *** shiftdiv: argument not a power of 2
endif
	n = 2 * n
	SHR reg,1
	ENDM		; of REPT
	ENDM
;****************************************************************************
scanaddr MACRO
	LOCAL	ADD_LOOP, ADD_DONE
;
; gets BX = scan line number, adds offset to DI
;
ifdef XHI
	ADD	DI,DI			; fix up col
endif
if scan_blks gt 1
	PUSH	BX
	PUSH	CX
	MOV	CX,SCAN_BLKS - 1
	AND	CX,BX
	JZ	ADD_DONE
	SUB	BX,CX		; now bx has last bits==0
ADD_LOOP: ADD	DI,BLK_DIFF
if scan_blks gt 2
	LOOP	ADD_LOOP
endif
ADD_DONE:
	POP	CX
endif
;
	PUSH	DX
if (scan_diff mod scan_blks) eq 0
	MOV	AX,SCAN_DIFF/SCAN_BLKS
else			; scan_diff mod scan_blks ne 0
	shiftdiv BX,SCAN_BLKS	; Here we assume SCAN_BLKS is a power of 2	
	MOV	AX,SCAN_DIFF
endif
	MUL	BX
	POP	DX
	ADD	DI,AX
;
if scan_blks gt 1
	POP	BX
endif
;
	ENDM
;****************************************************************************
row2scan MACRO	reg
;
; Receives reg = row. Sets BX = scan line
;
if CH_VSIZE EQ 8
ifdifi <reg>,<bl>
	MOV	BL,reg
endif
	XOR	BH,BH
	SHL	BX,1
	SHL	BX,1
	SHL	BX,1
else
	MOV	AL,CH_VSIZE
	MUL	reg
	MOV	BX,AX
endif
	ENDM
;****************************************************************************
incscan	MACRO	reg
	LOCAL	INC_BACK,INC_DONE
;
if scan_blks gt 1
	CMP	reg,BLK_DIFF*(SCAN_BLKS-1)
	JAE	INC_BACK
	ADD	reg,BLK_DIFF	; move on to next block
	JMP	INC_DONE
INC_BACK:	
endif
	ADD 	reg,SCAN_DIFF - BLK_DIFF*(SCAN_BLKS-1)
INC_DONE:
;
	ENDM
;****************************************************************************
decscan	MACRO	reg
	LOCAL	DEC_FRWD,DEC_DONE
if scan_blks gt 1
	CMP	reg,BLK_DIFF
	JB	DEC_FRWD
	SUB	reg,BLK_DIFF
	JMP	DEC_DONE
DEC_FRWD:
endif	
	SUB	reg,SCAN_DIFF - BLK_DIFF*(SCAN_BLKS-1)
DEC_DONE:
;
	ENDM
;****************************************************************************
invert	MACRO
ifdef INV
ifdef XHI
	XOR	WORD PTR ES:[DI],0FFFFH
else
	XOR	BYTE PTR ES:[DI],0FFH
endif
endif
	ENDM
;****************************************************************************
transfer MACRO	instr,reg
	LOCAL	T_LOOP
T_LOOP:	
ifdef EGA
	MOV	reg,ES:[DI]	; get into registers
endif
	MOV	reg,[SI]
	invert
	instr	ES:[DI],reg
	invert
ifdef XHI
	ADD	SI,4
else
	INC	SI
endif
	incscan	DI
	LOOP	T_LOOP	
	ENDM
;****************************************************************************
setchar MACRO	instr
	LOCAL	S_RET, S_OK1, S_OK2, S_OK3, S_LOOP
;
        PUSH    BP
;
	DEC	DI	; we get all coordinates counted from 1.
	DEC	BX	; This WILL change eventually
	DEC	CX
	DEC	DX	
;
	CMP	DI,MAX_HPOS
	JG	S_RET		; col > max. horiz. position ? return
	OR	DI,DI
	JL	S_RET		; col < 0 ? return
;
	MOV	BP,2*CH_VSIZE	; BP will contain the number of scan lines
	CMP	BX,DX		; clipping on top?
	JL	S_RET		; char bottom < screen top => leave
	JG	S_OK1		; char bottom > screen top => fits completely
	MOV	BP,CH_VSIZE	; must skip top half of character
	ADD	SI,BP		; increment source offset
ifdef XHI
	ADD	SI,BP
	ADD	SI,BP
	ADD	SI,BP
endif
	JMP	S_OK2		; skip over BX decrement
S_OK1:	DEC	BX		; BX must point to top position
S_OK2:	CMP	BX,CX		; clip on bottom?
	JG	S_RET		; char top > screen bottom => leave
	JL	S_OK3		; char top < screen bottom => fits completely
	MOV	BP,CH_VSIZE	; must skip bottom part only
;
S_OK3:	
	row2scan BL		; scan line in BX
	scanaddr	       	; offset added to DI
	MOV	CX,BP		; cx = #scan lines to display
;
ifdef XHI
	PUSH	SI
	PUSH	DI
	PUSH	CX
	transfer instr,AX
	POP	CX
	POP	DI
	POP	SI
	ADD	SI,2
	CALL	PAGE1
	transfer instr,AX
	CALL	PAGE0
else	
	transfer instr,AL
endif
;
S_RET:	POP     BP
	ENDM
;****************************************************************************
makecounts MACRO errlabel
;
; Places BX = #scan rows, CX = #columns (from BH,BL,CH,CL)
;
	SUB	CL,CH
	INC	CL		; ccount=c2-c1+1
	JLE	errlabel	; ccount <= 0 --return
	XOR	CH,CH	      	; CX = # cols to move
;
	SUB	BL,BH
	INC	BL		; rcount=r2-r1+1
	JLE	errlabel	; rcount <= 0 --return
	MOV	AL,CH_VSIZE	;
	MUL	BL
	MOV	BX,AX		; BX = #scans to move
;
	ENDM
;
endif				; of if1
;****************************************************************************
;
CODESEG SEGMENT PUBLIC 'CODE'
        ASSUME  CS:CODESEG
;
SCREENDR PROC	FAR
;
	PUSH	AX		; set up video segment in ES
	MOV	AX,VIDEO_SEG
	MOV	ES,AX
	POP	AX
;
        CMP     AH,0		; branch to appropriate service
        JNZ     SMORE1
        CALL    INIT
        RET
SMORE1: CMP     AH,1
        JNZ     SMORE2
        CALL    TERM
        RET
SMORE2: CMP     AH,2
        JNZ     SMORE3
        CALL    TEXTMODE
        RET
SMORE3: CMP     AH,3
        JNZ     SMORE4
        CALL    GRAPHIC
        RET
SMORE4: CMP     AH,4
        JNZ     SMORE5
        CALL    PUTROW
        RET
SMORE5: CMP     AH,5
        JNZ     SMORE6
        CALL    ORCHAR
        RET
SMORE6: CMP     AH,6
        JNZ     SMORE7
        CALL    XORCHAR
        RET
SMORE7: CMP     AH,7
        JNZ     SMORE8
        CALL    ERASE
        RET
SMORE8: CMP     AH,8
        JNZ     SMORE9
        CALL    MOVE
        RET
SMORE9: RET
SCREENDR ENDP
;
;-------------------------------------------------------------------
;

ifdef HERC
;
; port addresses
INDEX   EQU     03B4H
CNTRL   EQU     03B8H
CONFIG  EQU     03BFH
;
; control codes
SCRN_ON EQU     8
GRPH    EQU     2
TEXT    EQU     20H
;
GTABLE  DB      35H,2DH,2EH,07H
        DB      5BH,02H,57H,57H
        DB      02H,03H,00H,00H
TTABLE  DB      61H,50H,52H,0FH
        DB      19H,06H,19H,19H
        DB      02H,0DH,0BH,0CH
endif

ifdef TEC
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; The Tecmar manual refers to the "index register" as the POINTER
; REGISTER, so I renamed it PNTREG here. Similarly the control register
; is called the MODE SELECT REGISTER in the Tecmar manual as well as in
; the IBM technical reference manual, so I renamed it MSR.  Tecmar has
; an additional mode select register not found on the IBM graphics card,
; nor apparently the Hercules card.  It is called the EXTENDED MODE
; SELECT REGISTER ( XMSR ).  CSR refers to the COLOUR SELECT REGISTER
; which for some reason or other has to be set to a non zero value for
; this graphics mode, though not for some of the other Tecmar graphics
; modes.  Tecmar does not use the i/o address 03BFH.
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
;
; port addresses
PNTREG  EQU     03B4H
MSR     EQU     03B8H
XMSR    EQU     03BAH
CSR     EQU     03B9H
;
; control codes
SCRN_ON EQU     8      ;used to set the video on bit in the MSR
GRPH    EQU     18     ;number for MSR but with video bit = 0
TEXT    EQU     1      ;number for MSR but with video bit = 0
XTEXT   EQU     0      ;number for XMSR / text mode
XGRPH   EQU     50     ;number for XMSR / graphics mode
CSR_ON  EQU     1      ;sets colour select to a non-zero value for
                       ;graphics
;
GTABLE  DB      54,45,45,08   ;These numbers go to the 6845 CRT con-
        DB      91,01,86,88   ;troller's registers R0,R1,...,R11 in that
        DB      02,03,32,00   ;order, to set the 720x352 2-intensity
                              ;graphics mode.
;
TTABLE  DB      97,80,81,15   ;These numbers are used the same way to
        DB      25,04,25,25   ;set the 80x25 alphanumeric mode.
        DB      02,13,11,12
endif

;
;-------------------------------------------------------------------
;
;
INIT	PROC	NEAR

ifdef HERC
        MOV     DX,CONFIG
        MOV     AL,1
        OUT     DX,AL
endif
	CALL	TEXTMODE
	RET
INIT	ENDP
;
;-------------------------------------------------------------------
;
TERM	PROC	NEAR
	CALL	TEXTMODE
	RET
TERM	ENDP
;
;-------------------------------------------------------------------
;
GRAPHIC PROC    NEAR
;

ifdef GRMODE 			; use BIOS call to switch graphics mode
        MOV     AH,0
        MOV     AL,GRMODE
        INT     10H
endif

ifdef EGA
	MOV	AH,12H
	MOV	BL,10H	; get EGA information
	INT	10H
	CMP	BL,0	; installed memory
	JNZ	G_EGA128
	MOV	BYTE PTR CS:EGA64,1
G_EGA128:CMP	BH,0
	JZ	GR_COLOR
        MOV     AL,0FH
	JMP	GR_SET
GR_COLOR:MOV    AL,10H
GR_SET: MOV     AH,0
        INT     10H
endif

ifdef HERC
        PUSH    ES
        PUSH    DS
        MOV     AX,CS
        MOV     DS,AX
        MOV     AL,GRPH
        LEA     SI,GTABLE
        MOV     BX,0
        MOV     CX,4000H
        CALL    SETMD
        POP     DS
        POP     ES
endif

ifdef TEC
        PUSH    ES
        PUSH    DS
; Load the Color Select Register with a non-zero value
        MOV     DX,CSR
        MOV     AL,CSR_ON
        OUT     DX,AL
; Ready parameters for a call to SETMD
        MOV     AX,CS
        MOV     DS,AX
        MOV     AL,GRPH
        MOV     AH,XGRPH
        LEA     SI,GTABLE
        MOV     BX,0         ; Zero to blank screen
        MOV     CX,4000H     ; 720d x 352d / 8 < 8000H bytes
                             ; = 4000H words.
        CALL    SETMD
        POP     DS
        POP     ES
endif

ifdef WYSE
ifdef XHI
	MOV	DX,3DFH
	MOV	AL,0C8H
	OUT	DX,AL
;
else
	MOV	DX,3DFH
	MOV	AL,088H
	OUT	DX,AL
endif
;
	MOV	DX,3DDH
	XOR	AL,AL
	OUT	DX,AL
	INC	DX
	OUT	DX,AL
	MOV	AX,VIDEO_SEG
	MOV	ES,AX
ifdef INV
	MOV	AX,0FFFFH
else
	XOR	AX,AX
endif
	XOR	DI,DI
	MOV	CX,SC_VSIZE*(SC_HSIZE/16)
	REP	STOSW		; clear graphic screen
ifdef XHI
	CALL	PAGE1
	XOR	DI,DI
	MOV	CX,SC_VSIZE*(SC_HSIZE/16)
	REP	STOSW		; clear graphic screen
	CALL	PAGE0
endif
endif

        RET
;
ifdef EGA
EGA64	DB	0
endif
;
GRAPHIC ENDP
;
;-------------------------------------------------------------------
;
TEXTMODE    PROC    NEAR

ifdef HERC
        PUSH    ES
        PUSH    DS
        MOV     AX,CS
        MOV     DS,AX
        MOV     AL,TEXT
        LEA     SI,TTABLE
        MOV     BX,720H
        MOV     CX,2000
        CALL    SETMD
        mov     ah,2         ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        xor     dx,dx        ; I don't know why, but the text mode doesn't
        xor     bh,bh        ; seem to work without this interrupt.
        int     10h          ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        POP     DS
        POP     ES
else

ifdef TEC
        PUSH    ES
        PUSH    DS
; Set the color select register to zero
        MOV     DX,CSR
        XOR     AL,AL
        OUT     DX,AL
; Ready parameters for the call to SETMD
        MOV     AX,CS
        MOV     DS,AX
        MOV     AH,XTEXT
        MOV     AL,TEXT
        LEA     SI,TTABLE
        MOV     BX,720H      ; attribute byte + blank,
        MOV     CX,2000      ; count for clearing screen
        CALL    SETMD
        mov     ah,2         ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        xor     dx,dx        ; I don't know why, but the text mode doesn't
        xor     bh,bh        ; seem to work without this interrupt.
        int     10h          ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        POP     DS
        POP     ES
else

        MOV     AH,0
        MOV     AL,3
        INT     10H

endif
endif

        RET
TEXTMODE    ENDP
;

ifdef HERC
;-------------------------------------------------------------------
;
SETMD   PROC    NEAR
        PUSH    DS
        PUSH    ES
        PUSH    AX
        PUSH    BX
        PUSH    CX
        MOV     DX,CNTRL
        OUT     DX,AL
        MOV     AX,DS
        MOV     ES,AX
        MOV     DX,INDEX
        MOV     CX,12
        XOR     AH,AH
PARMS:  MOV     AL,AH
        OUT     DX,AL
        INC     DX
        LODSB
        OUT     DX,AL
        INC     AH
        DEC     DX
        LOOP    PARMS
        POP     CX
        MOV     AX,0B000H
        CLD
        MOV     ES,AX
        XOR     DI,DI
        POP     AX
        REP     STOSW
        MOV     DX,CNTRL
        POP     AX
        ADD     AL,SCRN_ON
        OUT     DX,AL
        POP     ES
        POP     DS
        RET
SETMD   ENDP
endif

ifdef TEC
;-------------------------------------------------------------------
;
SETMD   PROC    NEAR
        PUSH    DS
        PUSH    ES
        PUSH    AX
        PUSH    BX
        PUSH    CX
; Send number to MSR
        MOV     DX,MSR
        OUT     DX,AL
; Send number to XMSR
        MOV     DX,XMSR
        MOV     AL,AH
        OUT     DX,AL
; Ready parameters for the loop that loads the registers of the 6845
        MOV     AX,DS
        MOV     ES,AX
        MOV     DX,PNTREG
        MOV     CX,12
        XOR     AH,AH
        CLD
; The loop loading the registers of the 6845
PARMS:  MOV     AL,AH
        OUT     DX,AL
        INC     DX
        LODSB
        OUT     DX,AL
        INC     AH
        DEC     DX
        LOOP    PARMS
; Clear the screen
        POP     CX
        MOV     AX,0B000H
        MOV     ES,AX
        XOR     DI,DI
        POP     AX        ; AX now contains the value BX had at entry
        REP     STOSW
        MOV     DX,MSR
        POP     AX
; Set bit 3 in the MSR on to enable video
        OR      AL,SCRN_ON
        OUT     DX,AL
        POP     ES
        POP     DS
        RET
SETMD   ENDP
endif

;
;-------------------------------------------------------------------
;
PUTROW  PROC    NEAR
        CALL    PIXADDR 	; compute addr of (BX,CX) in ES:DI
        AND     CL,7
        MOV     CH,8
PRLOOP: CMP     DX,0
        JLE     PRRET
        CMP     DX,8
        JGE     PRSKIP
        MOV     CH,DL
PRSKIP: 
	PUSH    CX
        MOV     CL,CH
        MOV     BH,0FFH
        SHR     BH,CL		; BH = 0...01...1 with CH zeroes
	POP	CX
;
        MOV     AH,[SI]
ifdef INV
	XOR	AH,0FFH
endif
        XOR     AL,AL
        MOV     BL,0FFH
        ROR     AX,CL
        ROR     BX,CL
ifdef EGA
	NOT	BX
;
; higher byte first!!!
;
	PUSH	AX
	MOV	AL,BH
	CALL	WRMASK
	MOV	AH,ES:[DI]
	POP	AX
	MOV	ES:[DI],AH
	PUSH	AX
	MOV	AL,BL
	CALL	WRMASK
	MOV	AL,ES:[DI+1]
	POP	AX
        MOV     ES:[DI+1],AL
else
        AND     ES:[DI],BH      ; higher byte first!!!
        AND     ES:[DI+1],BL
        NOT     BX
        AND     AX,BX
        OR      ES:[DI],AH
        OR      ES:[DI+1],AL
endif
        SUB     DX,8
        INC     DI
        INC     SI
        JMP     PRLOOP
PRRET:  
ifdef EGA
	MOV	AL,0FFH
	CALL	WRMASK
endif
	RET
PUTROW  ENDP
;
;-------------------------------------------------------------------
;
; BH = 0...01...1 with CH zeroes
;
;
;
;-------------------------------------------------------------------
;
PIXADDR PROC    NEAR
;
; PURPOSE: Compute address from pixel coordinates
; RECEIVES: (BX,CX) = (pixel row, pixel col)
; RETURNS: ES:DI = screen address
;
	PUSH	AX
;
;
        MOV     DI,CX
        SHR     DI,1		; divide col number by 8
        SHR     DI,1
        SHR     DI,1
	scanaddr
;
	POP	AX
        RET
PIXADDR ENDP
;
;-------------------------------------------------------------------
;
ORCHAR 	PROC	NEAR

ifdef EGA
;
; set the writemode to OR
;
	MOV	AL,00010000B
	CALL	WRFUNC
;
ORSTART:
	setchar MOV
	XOR	AL,AL
	CALL	WRFUNC		; reset write mode
	RET
else 				; not EGA
	setchar	OR
	RET
endif

ORCHAR	ENDP
;
;-------------------------------------------------------------------
;
XORCHAR	PROC	NEAR

ifdef EGA
;
; set the writemode to XOR
;
	MOV	AL,00011000B
	CALL	WRFUNC
;
	JMP	ORSTART
else				; not EGA
	setchar	XOR
	RET
endif

XORCHAR	ENDP
;
;-------------------------------------------------------------------
;
MOVE	PROC	NEAR
;
ifdef EGA
	CMP	BYTE PTR CS:EGA64,0
	JZ	M_WRMOD1
	CMP	DL,0
	JZ	M_WRMOD1
	JMP	M_NOWR1
M_WRMOD1:MOV	AL,1
	CALL	WRMODE	
;
M_NOWR1:
endif
	CALL CLIP
; Now 0 <= r1, r2 <= max_hpos <= 127, 0 <= c1, c2 <= max_vpos <= 127
; Also assume -max_vpos <= dr <= max_vpos, -max_hpos <= dc <= max_hpos
;
	XOR	AL,AL
	SUB	AL,DH
	CMP	BH,AL		; r1 + dr < 0
	JGE	M_OK1		; no
	MOV	BH,AL		; yes-- r1 = -dr
M_OK1:	CMP	DH,0		; if dr < 0, then (max_vpos - dr) can be > 127
	JL	M_OK2		; but then r2 + dr < r2 <= max_vpos anyway
	MOV	AL,MAX_VPOS
	SUB	AL,DH
	CMP	BL,AL		; r2 + dr > maximum vertical?
	JLE	M_OK2		; no
	MOV	BL,AL		; yes-- r2 = max_vpos - dr
M_OK2:	XOR	AL,AL	
	SUB	AL,DL
	CMP	CH,AL		; c1 + dc < 0?
	JGE	M_OK3		; no
	MOV	CH,AL		; yes-- c1 = -dc
M_OK3:	CMP	DL,0		; if dc < 0, then (max_hpos - dr) can be > 127
	JL	M_OK4		; but then c2 + dc < c2 <= max_hpos anyway
	MOV	AL,MAX_HPOS
	SUB	AL,DL
	CMP	CL,AL		; c2 + dc > maximum horizontal?
	JLE	M_OK4		; no
	MOV	CL,AL		; yes-- c2 = max_hpos - dc
;
M_OK4:	CMP	DL,0
	JL	M_DCN		; dc negative
	STD			; dc positive--move downward
	MOV	AL,CL		; start from last column
	JMP	M_DCOK
M_DCN:	CLD			; move upward
	MOV	AL,CH		; start from 1st column
M_DCOK:	CMP	DH,0
	JL	M_DRN		; dr negative
	MOV	AH,BL		; start from last row + 1 
	INC	AH		; (we'll decrement the scan first)
	JMP	M_DROK
M_DRN:	MOV	AH,BH		; start from first row
;
M_DROK:	CALL	ADDR		; compute address of source 
	PUSH	ES
	POP	DS		; move from ES:DI to DS:SI
	MOV	SI,DI
	ADD	AH,DH
	ADD	AL,DL
	CALL	ADDR		; compute address of target
	makecounts M_RET
;
M_LOOP: 
	CMP	DH,0		; dr > 0?
	JL	M_UP		; if so, first move, then increment
	decscan	SI
	decscan	DI
M_UP:
	PUSH	SI
	PUSH	DI
	PUSH	CX

ifdef EGA
	CMP	BYTE PTR CS:EGA64,0
	JZ	M_NO_64
	CMP	DL,0
	JZ	M_NO_64
	CALL	M_64MOVE
	JMP	M_CONT
M_NO_64:
endif

ifdef XHI
	REP	MOVSW
else
	REP	MOVSB
endif
M_CONT:
	POP	CX
	POP	DI
	POP	SI

ifdef XHI
	CALL	PAGE1
	PUSH	SI
	PUSH	DI
	PUSH	CX
	REP	MOVSW
	POP	CX
	POP	DI
	POP	SI
	CALL	PAGE0
endif
;
	CMP	DH,0		; dr <= 0?
	JGE	M_DOWN		; if so, already decremented
	incscan	SI
	incscan	DI
;
M_DOWN:	DEC	BX
	JNZ	M_LOOP
;
M_RET:	CLD
ifdef EGA
	XOR	AL,AL
	CALL	WRMODE
endif
	RET
MOVE	ENDP
;
;-------------------------------------------------------------------
;
ifdef EGA
M_64MOVE PROC	NEAR
;
; Auxiliary to do the even/odd garbage in 64 K EGA's
;
; read a byte from plane 0 if even address, plane 1 if odd address
;
	PUSH	DX
M_64CONT:MOV	AL,4		; set read mode 0
	MOV	DX,3CEH
	OUT	DX,AL
	MOV	DX,3CFH
	MOV	AX,SI
	AND	AL,1		; set bit plane 
	OUT	DX,AL
	MOV	AL,DS:[SI]	; now byte in AL
	MOV	AH,ES:[DI]	; prime latches
	MOV	ES:[DI],AL	; write byte
;
	PUSHF
	POP	AX
	AND	AH,4		; direction flag
	JNZ	M_DECR
	INC	SI
	INC	DI
	JMP	M_INCR
M_DECR:	DEC	SI
	DEC	DI
M_INCR:	DEC	CX
	JNZ	M_64CONT
	POP	DX
	RET
M_64MOVE ENDP
endif
;
;-------------------------------------------------------------------
;
ERASE 	PROC	NEAR
;
	CALL	CLIP
;
	MOV	AH,BH
	MOV	AL,CH
	CALL	ADDR		; compute address in DI
	makecounts E_RET
;
	CLD
ifdef INV
	MOV	AX,0FFFFH
else
	XOR	AX,AX		; need AX for XHI only!
endif
;
E_LOOP:	PUSH	DI
	PUSH	CX
ifdef EGA
	MOV	AH,ES:[DI]	; get attention of registers
endif

ifdef XHI
	REP	STOSW
else
	REP	STOSB
endif
	POP	CX
	POP	DI

ifdef XHI
	CALL	PAGE1
	PUSH	DI
	PUSH	CX
	REP	STOSW
	POP	CX
	POP	DI
	CALL	PAGE0
endif

	incscan	DI
	DEC	BX
	JNZ	E_LOOP
;
E_RET:	RET
ERASE	ENDP
;
;-------------------------------------------------------------------
;
CLIP	PROC	NEAR
;
;
; we get all coordinates counted from 1.
; This WILL change eventually
;
	DEC	BH
	DEC	BL
	DEC	CH
	DEC	CL
;
	CMP	BH,0		; r1 < 0
	JGE	C_OK1		; no
	MOV	BH,0		; yes-- r1=0
C_OK1:	CMP	BL,MAX_VPOS	; r2 > max. screen row?
	JLE	C_OK2		; no
	MOV	BL,MAX_VPOS	; yes-- r2=nax. screen row
C_OK2:	CMP	CH,0		; c1 < 0?
	JGE	C_OK3		; no
	MOV	CH,0		; yes-- c1=0
C_OK3:	CMP	CL,MAX_HPOS	; c2 > max. screen col?
	JLE	C_OK4		; no
	MOV	CL,MAX_HPOS	; yes-- c2=max. screen col.
C_OK4:	RET
;
CLIP	ENDP
;
;-------------------------------------------------------------------
;
; ADDR computes the screen offset of (Chiwriter row-1, col-1) in (AH,AL)
; in DI
;
;
ADDR	PROC	NEAR
	PUSH	AX
	PUSH	BX
	MOV	DI,AX
	AND	DI,0FFH	
;
	row2scan AH
	scanaddr
	POP	BX
	POP	AX
	RET
ADDR	ENDP

;***************************************************************************
; EGA functions
;***************************************************************************
ifdef EGA
;
; Set the write function ( put, or, xor, and )
;
WRFUNC	PROC	NEAR
;
	PUSH	DX
	PUSH	AX
	MOV	AL,3
WROUT:	MOV	DX,3CEH
	OUT	DX,AL
	MOV	DX,3CFH
	POP	AX
	OUT	DX,AL
	POP	DX
	RET
WRFUNC	ENDP
;
; Set the write mode
;
WRMODE	PROC	NEAR
;
	PUSH	DX
	PUSH	AX
	MOV	AL,5
	JMP	WROUT
WRMODE	ENDP
;
;
; Set the write mode
;
WRMASK	PROC	NEAR
;
	PUSH	DX
	PUSH	AX
	MOV	AL,8
	JMP	WROUT
WRMASK	ENDP
endif

ifdef WYSE
ifdef XHI
PAGE0	PROC	NEAR
	PUSH	AX
	PUSH	DX
	MOV	DX,3DFH
	MOV	AL,0C8H
	OUT	DX,AL
	POP	DX
	POP	AX
	RET
PAGE0	ENDP
;
PAGE1	PROC	NEAR
	PUSH	AX
	PUSH	DX
	MOV	DX,3DFH
	MOV	AL,0CBH
	OUT	DX,AL
	POP	DX
	POP	AX
	RET
PAGE1	ENDP
endif
endif


;
CODESEG ENDS
	END
