/****************************************************************************
*
* chi2tex.c 
*
* ChiWriter -> TeX format conversion program
* (C) Copyright Cay S. Horstmann 1987. All Rights Reserved.
*
* This program may be freely copied for NON-COMMERCIAL purposes only.
* Written permission from the copyright holder is required if this 
* program or any part of it is given to someone in exchange for money 
* (even a "handling fee") or conditioned upon payment for a different 
* item ("bundling"). 
*
* DISCLAIMER OF LIABILITY. 
*
* The copyright holder does not warrant that this program does anything
* useful in particular or that it is free of adverse side effects. Any 
* similarities between the actual program behavior and the behavior 
* described in the associated documentation are merely a matter of 
* coincidence. ("It's free, so don't sue.")
*
* This program enables the use of the ChiWriter Scientific/Multifont word
* processor (Version 2.xx) as an aid to enter TeX files. 
*
* At this time (June '87), only special characters, fonts and simple super/sub-
* scripts are recognized and properly translated. It is hoped that TeX users
* familiar with the C programming language will extend this program to enable
* recognition of more structures. 
*
* It is possible to automatically produce those backslashes, braces...
* with a European keyboard by assigning them a different character. 
*
* Accented characters can also be automatically converted. 
*
* Marc Diener had the idea of using ChiWriter as a tool for entering TeX.
* This program was written by Cay Horstmann. 
*
* Please let us know of any ideas, suggestions, criticisms, etc...
* Our addresses are:
*
*     Marc Diener, Dept. of Mathematics, Univ. Paris VII, 2, pl Jussieu 45-55-5, 
*     75251 Paris Cedex 5, France
*
*     Cay S. Horstmann, Dept. of Mathematics and Computer Science, 
*     San Jose State Univ., San Jose, CA 95192-0103, USA
*
****************************************************************************/

/* General overview of the program.
** (1) We read in the conversion directives and store them.
** (2) We read in the font information in the ChiWriter file and match it
**     up against the information in the conversion directives. A translation
**     table mapping 
**         ChiWriter file font numbers <---> Conversion file font numbers
** 	 is built.
** (3) We skip the header/footer... blocks of the ChiWriter file.
**         If you want to grab more ChiWriter information (e.g. footnotes),
**     change the code here. 
** (4) We read in the ChiWriter document, one line at a time, into a buffer.
**     A ChiWriter line contains several super/subscripts. They are put into
**     a rectangular buffer (or rather 2 parallel ones, one for font, one
**     for character code.) 
**
**     +---------------------------------+
**     |   2     2             2i/n     |
**     | x   + y   = 1 if  = e      and | <-baseRow
**     |  i     i                        | <-lastRow
**     +---------------------------------+
**
**         If you want to recognize more ChiWriter features (e.g. centering,
**     hard page breaks, look here.
** (5) Only looking at the buffer rectangle, we convert and write a line at 
**     a time. 
**         If you want to add more recognition features here, you can 
**     change the code here without having to know about the ChiWriter
**     file format. 
**         Recommended addition: Look for horizontal and vertical 
**     combination symbols, e.g. 
**
**         ---> 
**              
**
*/
/* Syntax of Chiwriter Files
**
** \+ 		 superscript row follows
** \- 		 subscript row follows
** \= 		 end of text block
** 
** \Hx, \Fx  header/footer, x=D for default, E for even, 1..9 for page 1..9, 
** \S        separator
** \Nx       footnote text
** \Uxname   x=1..0, !..*, name = font name 
**           using font# x for the font with the given name
**
** \0 ... \9 font change (1-10)
** \! ... \* font change (11-20)
** \  		 soft space
** \, 		 hard return
** \/ 		 no/soft page break
** \\        backslash
** \@        page number
** \^        expanding marker (for centering)
** \[        tab
** \]x       reverse tab 
** \Fx       footnote
** 
*/

/*--------------------------------------------------------------------------*/

/* Syntax of Conversion Map file 
**
** Each line of the map file contains a command, starting with a 2 letter 
** command code. The code is followed by arguments, depending on the command.
**
** FO fontname font#
** 	to declare a font. The numbers should be between 1 and 20. 
**		They are used in other directives to avoid constantly writing out the full font 
**		name. If the numbers differ from the order of fonts in the ChiWriter
**    file, they are automatically rearranged.
**    e.g. FO ITALIC   3
**			  FO GREEK    7
**			  FO LINEDRAW 8
**         FO MATHII  10
**    STANDARD should always be 1. 
**
** CH font# char replacement
** 	e.g. CH 7 a \alpha
**		to replace all Greek "a" by "\alpha"
**    If the font number is 0, the replacement is done for all fonts. 
**
** FT font# On-command Off-command
**    to toggle a font on/off. 
**		e.g. FT 3 \it \rm
**
** To embed a blank into a TeX code word, use a  (ASCII 254). 
** Fascinating trivia fact: On some European keyboards, this code is produced
** when hitting the "umlaut" key and [Space]. 
**
** Suggestions for more commands:
**
** HO font1 char1 font2 char2 ... replacement
**    to translate a horizontal group of characters
**	   e.g. HORIZONTAL 8 - 8 - 8 - 8 L \arrow
**    would translate an ---> into the \arrow sequence.
**
** VE font1 char1 font2 char2 ... replacement
**    to translate a vertical group of characters
**    e.g. VERTICAL 10 i 10 j \int
*/

#include <stdio.h>

#define BOOL int
#define TRUE 1
#define FALSE 0

typedef char FILENAME[ 79 ]; 
	/* enough for 2 letters drive, 64 chars path, 8 file, 1 '.', 
	** 3 extension, 1 terminating 0
	*/

char Banner[] = "ChiWriter -> TeX Conversion. Version 1.00\n\
(C) Copyright Cay S. Horstmann 1987. All Rights Reserved.\n";

char Usage[]  = 
	"USAGE: chi2tex <input file> <output file> [-C<conversion directive file>]\n";

char Error[] = "*** ERROR ***";
char Warning[] = "*** WARNING ***";
char errOpen[] = "Can't open file:"; 
char warnFont[] = "File contains unknown font:";
char errMemory[] = "Memory overflow.";
char errFull[] = "Disk full.";

FILENAME cnvName = "CHI2TEX.CNV"; 
	/* name for map file */

FILE		*inFile;
FILE		*outFile;

/*--------------------------------------------------------------------------*/

/* attachExt( targetName, sourceName, extension )
** copies the source name into the target name. If no extension is present
** in the source name, the supplied extension is attached.
*/

attachExt( targetName, sourceName, extension )
FILENAME targetName, sourceName;
char *extension;

{	strcpy( targetName, sourceName );
	if( NULL == strrchr( sourceName, '.' ) )
		strcat( targetName, extension );
}

/*--------------------------------------------------------------------------*/

main( argc, argv )
int	argc;
char	*argv[];

{	FILE		*cnvFile;
	FILENAME inName;
	FILENAME outName;


	fprintf( stderr, Banner );

	if( argc < 3 ) { printf( Usage ); return(0); } 
		/* the turkey didn't even enter 2 file names */

	inFile = outFile = cnvFile = NULL; 

	while( --argc ) /* scan command line arguments */ 
	{	if( '-' == (*++argv)[ 0 ] ) /* option */ 
	 	{	switch( toupper( (*argv)[ 1 ] ) ) /* check option letter */
			{
			case 'C': 	attachExt( cnvName, *argv + 2, ".CNV" );
							break;

			/* insert more options here as they come up. 
			*/

			default: 	printf( Usage ); return(0); /* funny option */
			}
		}
		else if( !inFile ) /* first file name is input file */ 
		{	attachExt( inName, *argv, ".CHI" );
			if( NULL == ( inFile = fopen( inName, "r" )))
			{	printf( "%s %s %s\n", Error, errOpen, inName ); 
				return( 0 );
			}
		}
		else if( !outFile ) /* next one is output file */
		{	attachExt( outName, *argv, ".TEX" );
			if( NULL == ( outFile = fopen( outName, "w" )))
			{	printf( "%s %s %s\n", Error, errOpen, outName ); 
				return( 0 );
			}
		}
		else 	{ printf( Usage ); return( 0 ); } /* what? another file name? */
	} /* while */

	if( !inFile || !outFile ) { printf( Usage ); return( 0 ); } 

	if( NULL == ( cnvFile = fopen( cnvName, "r" )))
	{	printf( "%s %s %s\n", Error, errOpen, cnvName ); 
		return( 0 );
	}
	if( !readMap( cnvFile ) ) 
	{ 	printf( "%s %s\n", Error, errMemory ); 
		return( 0 ); 
	}
	fclose( cnvFile );

	chi2tex();

	fclose( inFile );
	fclose( outFile );

	return( 1 );
}
		
/*--------------------------------------------------------------------------*/

/* We do our own memory allocation here since we are unsure how good the
** malloc() of the C package is. We never have to free space and don't
** like to waste the overhead involved, as we will be requesting many
** blocks to store very short strings. 
*/

#define SSSIZE 32000
unsigned char stringSpace[ SSSIZE ];
#define SPOINTER unsigned
SPOINTER sfree = 1;

#define FAKESPACE 254
/* this is a '' which is used to denote a space inside a replacement string */

/*--------------------------------------------------------------------------*/

SPOINTER storeString( s )
unsigned char *s;

/* Store the string s, starting at loc and return the next free location
** thereafter. If memory overflow, return 0.
*/

{	unsigned char *p;
	unsigned ch;
	SPOINTER start;

	start = sfree;
	p = stringSpace + sfree;
	while( ch = *s++ )
	{	if( sfree++ >= SSSIZE ) return( 0 );
		*p++ = ch == FAKESPACE ? ' ' : ch;
	}
	*p = 0; 
	sfree++;
	return( start );
}

/*--------------------------------------------------------------------------*/

#define NFONT 20
/* maximum number of fonts in the ChiWriter file */
#define FNAMELEN 8
/* A font name has <=8 characters */

char fontName[ NFONT + 1 ] [ FNAMELEN + 1 ];
/* Array with the names of the fonts. Entry 0 is not used. */

#define NCHAR 256

/* if storage gets tight, can compactify into the 150-odd ASCII + extended
** characters really used.
*/

typedef struct
{	SPOINTER C_REPL[ NCHAR ]; 
		/* points to the replacement text of the character */
} CHARDATA;

typedef struct
{	SPOINTER F_ON;
	SPOINTER F_OFF;
	CHARDATA *F_CHAR; 
		/* info for characters in this font. allocated as needed. */
} FONTDATA; 

FONTDATA fontInfo[ NFONT ];

extern unsigned char *calloc(); 

/*--------------------------------------------------------------------------*/

unsigned char *foundRemap( ft, ch )
unsigned ft, ch;

{	CHARDATA *pc;
	SPOINTER offset;

	if( ( pc = fontInfo[ ft ].F_CHAR ) && (offset = pc->C_REPL[ ch ] ) )
		return( stringSpace + offset );
	else if( ( pc = fontInfo[ 0 ].F_CHAR ) && (offset = pc->C_REPL[ ch ] ) )
		return( stringSpace + offset );
	else return( NULL );
}

/*--------------------------------------------------------------------------*/

unsigned char *foundToggle( ft, on )
unsigned ft;
BOOL on;

{	SPOINTER offset;

	if( offset = on ? fontInfo[ ft ].F_ON : fontInfo[ ft ].F_OFF ) 
		return( stringSpace + offset );
	else
		return( NULL );
}

/*--------------------------------------------------------------------------*/

readMap( cnvFile )
FILE *cnvFile;

{	unsigned char Buffer[ 80 ];
	unsigned char TeXT1[ 80 ], TeXT2[ 80 ];
	unsigned char fname[ FNAMELEN+1 ];
	unsigned char ch[ 2 ];
	unsigned font;
	CHARDATA *pc;

	do
 	{	fgets( Buffer, sizeof(Buffer), cnvFile); /* read one line */
		if( !strncmp( Buffer, "FO", 2 ) ) 
		{	sscanf( Buffer+2, "%d %9s", &font, fname );
			if( font < NFONT ) strcpy( fontName[ font ], fname );
		}
  		else if( !strncmp( Buffer, "CH", 2 ) ) 
		{	sscanf( Buffer+2, "%d %1s %s", &font, ch, TeXT1 ); 
			if( !( pc = fontInfo[ font ].F_CHAR ) 
				&& !( pc = fontInfo[ font ].F_CHAR = (CHARDATA *)calloc( sizeof( CHARDATA ), 1 ) ) )
				return( FALSE );
			if( !( pc->C_REPL[ *ch ] = storeString( TeXT1 ) ) ) 
				return( FALSE );
		}
		else if( !strncmp( Buffer,"FT", 2 ) ) 
		{	sscanf( Buffer+2, "%d %s %s", &font, TeXT1, TeXT2 );
			if( !( fontInfo[ font ].F_ON = storeString( TeXT1 ) ) 
				|| !( fontInfo[ font ].F_OFF = storeString( TeXT2 ) ) )
				return( FALSE );
		}
	} while( !feof( cnvFile ) );
	
	return( TRUE );
}

/*--------------------------------------------------------------------------*/

#define BUFLENGTH 250
char Buffer[ BUFLENGTH ];
BOOL inEof, outFull;

unsigned inFont, outFont;

int baseRow;
int lastRow;

#define NROWS 10
#define NCOLS 80
unsigned char lineChar[ NROWS ] [ NCOLS ];
unsigned char lineFont[ NROWS ] [ NCOLS ];

unsigned char fontRemap[ NFONT + 1 ];

BOOL spaceWarn;

/*--------------------------------------------------------------------------*/

getline() 

{	inEof |= fgets( Buffer, BUFLENGTH, inFile ) == NULL;
} 

/*----------------------------------------------------------------------*/

chi2tex()

{	char 	ch;
	char	fname[ FNAMELEN+1 ];
	BOOL	done;
	unsigned font, f;
	
	inEof = outFull = done = FALSE; 
	SETMEM( fontRemap, sizeof( fontRemap), 0 );
	getline();

	if( !strncmp( Buffer, "\\1cw", 4 ) ) getline(); 
		/* Version 2 signature */
		/* recall that strncmp == 0 iff the strings are equal */

/* NOTE: At this point we are skipping all headers/footers/footnotes. 
** Ambitious programmers may want to convert them to TeX format.
*/

	while(!inEof && !outFull && !done)
		if( Buffer[ 0 ] == '\\' )
			switch( Buffer[ 1 ] )
			{
			case 'U': 
       		font = Buffer[2]; 
	         if( '1' <= font && font <= '9' ) font -= '0';
				else if( font == '0' ) font = 10;
				else if( '!'<= font && font <='*' ) font -= '!'-11;
	         else font = 0;

	         sscanf( Buffer + 3, "%s", fname );
	         for( f=1; f<=NFONT; f++) /* search for font in the font list */
	            if( !strcmp( fname, fontName[ f ] ) )
	            {  fontRemap[font]=f; break; }
				if( f > NFONT ) 
					printf("%s %s %s\n",Warning, warnFont, fname );
				getline();
				break;
			case 'S': 
				skipBlock(); break; 	/* skip separator */
			case 'H':
				skipBlock(); break; 	/* skip header */
			case 'F': 
				skipBlock(); break; 	/* skip footer */
			case 'N':
				skipBlock(); break; 	/* skip footnote */
			default:	 	
				readBlock(); 			/* read and convert main text block */
		 		done = TRUE;
				break;
			} /* switch */
		else
			getline(); /* this should never happen!!! */
}

/*----------------------------------------------------------------------*/

skipBlock()

{	while( !inEof && strncmp( Buffer, "\\=", 2 ) ) getline(); 
		/* recall that strncmp == 0 iff the strings are equal */
	getline(); 
}

/*----------------------------------------------------------------------*/

readBlock( )

{	unsigned row; 

	inFont = outFont = 1; 
	
	while( !inEof )
	{	if( !strncmp( Buffer, "\\=", 2 ) ) { getline(); return; }
	
		row = 0; 
		while( !strncmp( Buffer, "\\+", 2 ) ) 
			readRow( row++, FALSE /* isBase */ );
		baseRow = row; 
		readRow( row++, TRUE /* isBase */  );
		while( !strncmp( Buffer, "\\-", 2 ) ) 
			readRow( row++, FALSE /* isBase */ );
		lastRow = row - 1;

		writeLine();
	}
}

/*----------------------------------------------------------------------*/

readRow( row, isBase )
int row;
	/* row of (lineChar,lineFont) into which the text row should be placed.
BOOL isBase; 
	/* TRUE if reading base row, FALSE if reading super/subscript row */

{	unsigned ch;
	int length, col;
	
   if( inEof || row >= NROWS ) return;

	SETMEM( lineChar[ row ], NCOLS, 0 );
	SETMEM( lineFont[ row ], NCOLS, 0 );
	
	col = length = 0; 
	while( ( ch = Buffer[ col++ ] ) && length < NCOLS )
	{	if( ch == '\\' )
		{	ch = Buffer[ col++ ]; 
			if( '0' <= ch && ch<='9' || '!' <= ch && ch <= '*' )
			{	inFont = '1' <= ch && ch <= '9' ? ch - '0' : 
					ch == '0' ? 10 : ch - '!' + 11;
            if( fontRemap[ inFont ] ) inFont = fontRemap[ inFont ];
         }
			else if(ch=='\\') 
			{ 	lineChar[ row ][ length ] = '\\'; 
				lineFont[ row ][ length++ ] = inFont;
			}
			else if(ch==' ')  length++; /* soft space */

/* All the following features are just ignored.
** Some of them could well be translated into TeX, 
** in particular hard page breaks, centering and footnotes.
*/
			else if(ch==',')  ; /* hard return */
			else if(ch=='/')  ; /* soft page break */
			else if(ch=='@')  length++; /* page number */
			else if(ch=='[')  length++; /* tab marker */
			else if(ch=='^')  length++; /* expanding filler */
			else if(ch==']')  { length++; col++; } /* reverse tab */ 
			else if(ch=='F')  { length++; col++; } /* footnote */
		}
		else if( ch==' ' )  lineChar[ row ][ length++ ] = isBase ? ' ' : 0;
		else if( ch>= '!') /* character key */ 
		{ 	lineChar[ row ][ length ] = ch; 
			lineFont[ row ][ length++ ] = inFont; 
		}
	}
	getline(); /* get next line */
}

/*--------------------------------------------------------------------------*/

writeLine()

/* Convert and write the line assembled in the line buffer
*/

{	int col;
	unsigned ch, font;
	unsigned char *s;
	int superSub; /* -1 for super, 0 for base, 1 for sub */
	int brace; /* number of open braces */
	unsigned row;
#define FSSIZE 1
/* just one open level of TeX groups {...} right now */
	unsigned fontStack[ FSSIZE ]; 

	spaceWarn = FALSE;

	col = 0; brace = 0; superSub = 0; row = baseRow;
	do
	{	if( ch = lineChar[ row ][ col ] ) 
		{		/* find whether character should be remapped. */
			if( s = foundRemap( font = lineFont[ row ][ col ], ch ) )
				sendString( s );
			else 
			{	/* Is the font different from the current font?
				** If so, was the current font special? Then turn it off?
				** Is the new font special? Then turn it on. 
				*/

				if( font && font != outFont )
				{	if( s = foundToggle( outFont, 0 ) ) sendString( s );
					outFont = font;
					if( s = foundToggle( outFont, 1 ) ) sendString( s );
				}
				sendChar( ch ); 
			}
		}
		else 	
		{	if( !superSub ) 	/* check super/subscript */
			{	if( row >= 1 && ( ch = lineChar[ row-1 ][ col ] ) )
				{	sendChar( '^' ); superSub = -1; }
				else 
				if( row < lastRow && ( ch = lineChar[ row+1 ][ col ] ) )
				{	sendChar( '_' ); superSub = 1; }

				if( superSub) /* found one */
				{	row += superSub; 
					if( lineChar[ row ][ col + 1] ) 
					{ fontStack[ brace++ ] = outFont; sendChar( '{' ); }
						/* super/subscript longer than 1 char */

					col--; /* force reinspection of the column */
				}
			}
			else /* are currently in super/subscript */
			{	if( brace ) { outFont = fontStack[ --brace ]; sendChar( '}' ); }
				row -= superSub; 
				superSub = 0;
				col--; /* reinspect base row here */
			}
		}
	}	while( ++col < NCOLS );

	if( brace ) { outFont = fontStack[ --brace ]; sendChar( '}' ); }
	sendChar( '\n' );
}

/*--------------------------------------------------------------------------*/

isletter( x )
unsigned x;

{	return( 'A' <= x && x <= 'Z' || 'a' <= x && x <= 'z' );
}


/*--------------------------------------------------------------------------*/

/* The main problem now is when to insert additional spaces into the
** TeX output stream. 
**
** A space is needed after a control word, i.e. a string starting with
** \ and containing letters. 
** 
** sendString checks whether its argument is a control word.
** If so, it sets spaceWarn.
**
** If spaceWarn is set and another letter is to be sent, a space
** is sent first. 
*/

sendString( s )
unsigned char *s;

{	unsigned ch;
	BOOL insideWord;

	insideWord = FALSE;
	while( ch = *s++ )
	{	sendChar( ch );
		if( ch == '\\' ) insideWord = TRUE; 
		else if( insideWord ) insideWord = isletter( ch );
	}
	spaceWarn = insideWord; 
}

/*--------------------------------------------------------------------------*/

sendChar( ch )
unsigned ch;

{	if( spaceWarn ) 
	{	spaceWarn = FALSE;
		if( isletter( ch ) ) sendChar(' ');
		/* recursive--be careful to reset spaceWarn first! */
	}
	
	if( !outFull ) 
		if( outFull = fputc( ch, outFile ) == EOF )
			printf("%s %s\n",Error, errFull );
}
